//======================================================== IBJprt.c
//
// IBJparticle: Lagrangian Particle Dispersion Model
// =================================================
//
// Copyright (C) Dr. Lutz Janicke, Dunum, Germany, 1999-2000
// email: lj@janicke.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// history:
//
// 98-07-31  lj  1.01
// 98-08-31  lj  1.02   reference to MP and NP removed
// 98-09-06  lj  1.03   statistical variation of first time step
// 99-05-03  lj  1.5.1  use of modified random number generator
// 99-09-01  uj  1.5.2  implementation of updated IBJ routines
//
//==================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include <signal.h>

#include "IBJmsg.h"
#include "IBJtxt.h"
#include "IBJary.h"
#include "IBJrnd.h"
#include "IBJdmn.h"
#include "IBJmth.h"
#include "IBJgrd.h"
#include "IBJprm.h"
#include "IBJp3d.h"
#include "IBJtbl.h"
#include "IBJprt.h"

#define  VERSION  "1.5.2"

#define VX3(a,b,c)  (*(float*)AryPtrX(&VxDsc,(a),(b),(c)))
#define VY3(a,b,c)  (*(float*)AryPtrX(&VyDsc,(a),(b),(c)))
#define VS3(a,b,c)  (*(float*)AryPtrX(&VsDsc,(a),(b),(c)))
#define WW3(a,b,c)  (*(VECTOR*)AryPtrX(&WwDsc,(a),(b),(c)))
#define EE3(a,b,c)  (*(MATRIX*)AryPtrX(&EeDsc,(a),(b),(c)))

/*==========================================================================*/

static char *eMODn = "IBJprt";

float Zmax;
float C0, A1, B1, C1;
int PrtMode, PrtCheck;
double T1, T2;
static int Ticks, AddRandom, AddWrite;

ARYDSC  ConDsc;
#define CON(a,b,c)  (*(double*)AryPtrX(&ConDsc,(a),(b),(c)))
ARYDSC  IcpDsc;
#define ICP(a,b,c)  (*(double*)AryPtrX(&IcpDsc,(a),(b),(c)))
ARYDSC  DryDsc;
#define DRY(a,b)    (*(double*)AryPtrX(&DryDsc,(a),(b)))
ARYDSC  WetDsc;
#define WET(a,b)    (*(double*)AryPtrX(&WetDsc,(a),(b)))
ARYDSC  TmpDsc;
#define TMP2(a,b)   (*(float*)AryPtrX(&TmpDsc,(a),(b)))
#define TMP3(a,b,c) (*(float*)AryPtrX(&TmpDsc,(a),(b),(c)))

ARYDSC  VvDsc;
#define vVV(a)  (*(VECTOR*)AryPtrX(&VvDsc,(a)))
ARYDSC  WwDsc;
#define vWW(a)  (*(VECTOR*)AryPtrX(&WwDsc,(a)))
ARYDSC  SsDsc;
#define mSS(a)  (*(MATRIX*)AryPtrX(&SsDsc,(a)))
ARYDSC  LlDsc;
#define mLL(a)  (*(MATRIX*)AryPtrX(&LlDsc,(a)))
ARYDSC  KkDsc;
#define mKK(a)  (*(MATRIX*)AryPtrX(&KkDsc,(a)))
ARYDSC  PpDsc;
#define mPP(a)  (*(MATRIX*)AryPtrX(&PpDsc,(a)))
ARYDSC  EeDsc;
#define mEE(a)  (*(MATRIX*)AryPtrX(&EeDsc,(a)))

#define BP2(a,b)  (*(float*)AryPtrX(&(pGrd->bpdsc),(a),(b)))
#define BM2(a,b)  (*(float*)AryPtrX(&(pGrd->bmdsc),(a),(b)))
#define PD2(a,b)  (*(float*)AryPtrX(&PdDsc,(a),(b)))
#define LI3(a,b,c)  (*(float*)AryPtrX(&LiDsc,(a),(b),(c)))

PTLRECPTR PtlTab;

FILE *CmdFile, *ChkFile;
char CmdName[MAXFNLEN+1];
char LogName[MAXFNLEN+1];
char ChkName[MAXFNLEN+1];
char WorkPath[MAXFNLEN+1];
float W2p, Tdep;
int Nptl;
double Tm;

static int ClcModField( void );        /* Berechnung des Modell-Feldes         */

int ClcModValue(                /* Feld am Ort des Teilchens            */
LOCREC *pl,
MODREC *pm );

int ClcLoc(    /* Feststellen, in welcher Masche das Teilchen ist */
VECTOR *pp,
LOCREC *pl );

PTLREC *NewPtl( void );         /* Neues Teilchen zur Verfgung stellen */

int InitPtl( PTLREC *pp )       /* Teilchen-Geschwindigkeit initialisieren */
  ;
int GetPtl( void );             /* Teilchen erzeugen                    */

PTLREC *FirstPtl( void );       /* Pointer auf erstes Teilchen          */

PTLREC *NextPtl( PTLREC *pp );  /* Pointer auf nchstes Teilchen        */

static int AddDose(             /* Beitrag zur Dosis aufaddieren        */
PTLREC *pa, LOCREC *pold,       /* Alte Position                        */
PTLREC *pe, LOCREC *pnew,       /* Neue Position                        */
int flag )                      /* Kennzeichnung des Schrittes          */
  ;
int RunPtl( void );             /* Serie von Zeitschritten durchfhren  */

int ClcCon( void );             /* Berechnung der Konzentration         */


/*============================================================== MakeFileName
*/
char *MakeFileName( char *name, char *extension )
  {
  dP(MakeFileName);
  static char fname[MAXFNLEN+1];
  char *pc, *ps, *pe, n[MAXFNLEN+1];
  if (!strncmp(name, "LOG", 3))  return name;
  if (strlen(name) > MAXFNLEN)                          eX(1);
  strcpy(n, name);
  for (pc=n; (*pc); pc++)  if (*pc == ALTPSEP)  *pc = PATHSEP;
  ps = strrchr(n, PATHSEP);
  pe = strrchr(n, '.');
  if ((ps) && (pe) && (pe < ps))  pe = NULL;
  if (ps)  *fname = 0;
  else  strcpy(fname, WorkPath);
  if (strlen(fname) + strlen(n) > MAXFNLEN)             eX(2);
  strcat(fname, n);
  if (!pe) {
    if (strlen(fname) + strlen(extension) > MAXFNLEN)   eX(3);
    strcat(fname, extension);
    }
  return fname;
eX_1:  eX_2:  eX_3:
  *fname = 0;
  vMsg("maximum length of file name exceeded! (%s)\n", ePGMs);
  return fname;
  }

/*================================================================== DefineGrid
*/
static int DefineGrid( void )
  {
  dP(DefineGrid);
  int k;
  if (PrtMode & PRT_DEFGRD)                     eX(0);
  PrtMode |= PRT_DEFGRD;
  if      (strstr(OP, "-perx"))     PrtMode &= ~PRT_PEX;
  else if (strstr(OP,  "perx"))     PrtMode |=  PRT_PEX;
  if      (strstr(OP, "-pery"))     PrtMode &= ~PRT_PEY;
  else if (strstr(OP,  "pery"))     PrtMode |=  PRT_PEY;
  if      (strstr(OP, "-nodrift"))  PrtMode &= ~PRT_NODRIFT;
  else if (strstr(OP,  "nodrift"))  PrtMode |=  PRT_NODRIFT;
  if      (strstr(OP, "-met3"))     PrtMode &= ~PRT_3DM;
  else if (strstr(OP,  "met3"))     PrtMode |=  PRT_3DM;
  if      (strstr(OP, "-velcor"))   PrtMode &= ~PRT_VELCOR;
  else if (strstr(OP,  "velcor"))   PrtMode |=  PRT_VELCOR;
  W2p = sqrt(2*3.1415926);
  if (NZ < 1)                                   eX(15);
  if (PRTMET3) {
    if (NX<1 || NY<1 || DD<=0)                  eX(11);
    X1 = X0 + NX*DD;
    Y1 = Y0 + NY*DD;
    }
  else {
    NX = 0;
    NY = 0;
    DD = 100;
    if (X1<=X0 || Y1<=Y0)                       eX(12);
    }
  if (DA <= 0) {
    DA = DD;
    A0 = X0;
    B0 = Y0;
    if (PRTMET3) {
      if (MX <= 0)  MX = NX;
      if (MY <= 0)  MY = NY;
      }
    else {
      if (MX <= 0)  MX = 1;
      if (MY <= 0)  MY = 1;
      }
    }
  A1 = A0 + MX*DA;
  B1 = B0 + MY*DA;
  if (MZ > 0) {
    if (CC(1) > 0) {
      C0 = CC(0);
      for (k=1; k<=MZ; k++)  if (CC(k) <= CC(k-1))      eX(13);
      }
    else {
      if (DC <= 0)                                      eX(14);
      CC(0) = C0;
      for (k=1; k<=MZ; k++)  CC(k) = C0 + k*DC;
      }
    }
  else  CC(0) = C0;
  C1 = CC(MZ);
  if (NZ < 1)                                           eX(10);
  Zmax = ZZ(NZ);
  if (ZZ(0))                                            eX(1);
  for (k=1; k<=NZ; k++)  if (ZZ(k) <= ZZ(k-1))          eX(2);
  if (PRTMET3) {
    if (A0<X0 || A1>X1 || B0<Y0 || B1>Y1)               eX(3);
    GrdDefK(0, X0, Y0, X1, Y1, DD, NZ, &ZZ(0), SH, ZH); eG(4);
    }
  AryCreate(&ConDsc, sizeof(double), 3, 1, MX, 1, MY, 1, MZ);    eG(20);
  AryCreate(&IcpDsc, sizeof(double), 3, 1, MX, 1, MY, 1, MZ);    eG(21);
  AryCreate(&DryDsc, sizeof(double), 2, 1, MX, 1, MY);           eG(22);
  AryCreate(&WetDsc, sizeof(double), 2, 1, MX, 1, MY);           eG(23);
  if (!PRTMET3) {
    AryCreate(&VvDsc, sizeof(VECTOR), 1, 0, NZ);          eG(24);
    AryCreate(&WwDsc, sizeof(VECTOR), 1, 1, NZ);          eG(25);
    AryCreate(&SsDsc, sizeof(MATRIX), 1, 0, NZ);          eG(26);
    AryCreate(&LlDsc, sizeof(MATRIX), 1, 0, NZ);          eG(27);
    AryCreate(&KkDsc, sizeof(MATRIX), 1, 0, NZ);          eG(28);
    AryCreate(&PpDsc, sizeof(MATRIX), 1, 0, NZ);          eG(29);
    AryCreate(&EeDsc, sizeof(MATRIX), 1, 0, NZ);          eG(30);
    }
  RndSet((long)(SD+AddRandom));
  return 0;
eX_0:
  eMSG("redefinition of grid not allowed! (%s)\n", ePGMs);
eX_1:  eX_2: eX_10: eX_11: eX_12: eX_15:
  eMSG("invalid computational grid! (%s)\n", ePGMs);
eX_3:
  eMSG("sampling grid larger than computational grid! (%s)\n", ePGMs);
eX_4:
  eMSG("grid not initialized! (%s)\n", ePGMs);
eX_13: eX_14:
  eMSG("invalid sampling grid! (%s)\n", ePGMs);
eX_20: eX_21: eX_22: eX_23:
eX_24: eX_25: eX_26: eX_27: eX_28: eX_29: eX_30:
  eMSG("can't allocate arrays! (%s)\n", ePGMs);
  }

/*================================================================ ClcModField
*/
static int ClcModField( void )        /* Berechnung des Modell-Feldes */
  {
  dP(ClcModField);
  int k;
  float tau, su, sv, sw, sc;
  float taum, taup;
  float s0, w2, wp, ga, fs, dz;
  VECTOR wind, del1, del2;
  MATRIX sigma, psi, phi, omega, rot, trot, kappa;
  MATRIX ident, denom, sigp, sigm, psip, psim, fak1, fak2;
  if (PRTMET3) {
    vMsg("calculating 3d-fields ...\n");
    PrtClc3dm();                                            eG(1);
    vMsg("... done\n");
    return 0;
    }
  vMsg("calculating 1d-fields\n");
  for (k=0; k<=NZ; k++) {
    if      (TU(k) > 0)  KU(k) = SU(k)*SU(k)*TU(k);
    else if (SU(k) > 0)  TU(k) = KU(k)/(SU(k)*SU(k));
    else TU(k) = 10;
    if      (TV(k) > 0)  KV(k) = SV(k)*SV(k)*TV(k);
    else if (SV(k) > 0)  TV(k) = KV(k)/(SV(k)*SV(k));
    else TV(k) = 10;
    if      (TW(k) > 0)  KW(k) = SW(k)*SW(k)*TW(k);
    else if (SW(k) > 0)  TW(k) = KW(k)/(SW(k)*SW(k));
    else TW(k) = 10;
    }
  if (TA > 0)                               /* Zeitschritt festlegen,    */
    for (k=0; k<=NZ; k++)  TT(k) = TA;          /* absolut oder ...      */
  else {
    if (AA > 0)
      for (k=0; k<=NZ; k++)  TT(k) = AA*TW(k);  /* ... relativ zu Tw ... */
    else
      for (k=0; k<=NZ; k++)  TT(k) = 1;         /* ... oder Default.     */
  }
  ident = DefM(1, 0, 0, 0, 1, 0, 0, 0, 1);
  for (k=0; k<=NZ; k++) {
    tau = TT(k);
    su  = SU(k);
    sv  = SV(k);
    sw  = SW(k);
    sc  = SC(k);
    wind  = DefV(VX(k), VY(k), 0.0);
    rot   = DefR2(VX(k), VY(k));
    trot  = TrnM(rot);
    sigma = DefM(su*su, 0, -sc*sc, 0, sv*sv, 0, -sc*sc, 0, sw*sw);
    sigma = DotMM(rot, DotMM(sigma, trot));
    kappa = DefM(KU(k), 0, 0, 0, KV(k), 0, 0, 0, KW(k));
    kappa = DotMM(rot, DotMM(kappa, trot));
    phi   = DotMM(sigma, InvM(kappa));
    denom = InvM (AddMM(ident, MulSM(0.5*tau, phi)));
    psi   = DotMM(SubMM(ident, MulSM(0.5*tau, phi)), denom);
    omega = SubMM(sigma, DotMM(psi, DotMM(sigma, TrnM(psi))));
    vVV(k) = wind;
    mSS(k) = sigma;
    mKK(k) = kappa;
    mPP(k) = psi;
    mEE(k) = ChlM(sigma);
    mLL(k) = ChlM(omega);
    }
  for (k=1; k<=NZ; k++) {
    dz = ZZ(k) - ZZ(k-1);
    taup = TT(k);  taum = TT(k-1);
    sigp = mSS(k);  sigm = mSS(k-1);
    psip = mPP(k);  psim = mPP(k-1);
    del1 = SubVV(DefV(taup*sigp.zx, taup*sigp.zy, taup*sigp.zz),
                 DefV(taum*sigm.zx, taum*sigm.zy, taum*sigm.zz));
    del2 = SubVV(DefV(sigp.zx, sigp.zy, sigp.zz),
                 DefV(sigm.zx, sigm.zy, sigm.zz));
    fak1 = MulSM(0.25/dz, AddMM(SubMM(ident,psip), SubMM(ident,psim)));
    fak2 = MulSM(0.5/dz,  AddMM(MulSM(taup, psip), MulSM(taum, psim)));
    vWW(k) = AddVV(DotMV(fak1,del1), DotMV(fak2,del2));
    }
  if (VD > 0) {
    s0 = SW(0);
    w2 = sqrt(2.0);
    wp = sqrt(Pi);
    ga = VS/(w2*s0);
    fs = wp*ga + exp(-ga*ga)/(2 - Erfc(ga));
    PD = 2*wp*VD/(w2*s0*fs + wp*VD);
    if (PD > 1) {
      vMsg("PWd=%1.2f !\n", PD);
      PD = 1;
      }
    }
  return 0;
eX_1:
  eMSG("can't calculate 3-dim meteorology! (%s)\n", ePGMs);
  }

/*==================================================================== Int3dv
*/
static int Int3dv(  /* Interpolate velocity field in 3 dimensions   */
LOCREC *pl,         /* location within grid                 */
VECTOR *pv )        /* interpolated field                   */
  {
  float a1, a0, b1, b0, c1, c0;
  int i, j, k;
  float v1, v0;
  i = pl->m.i;
  j = pl->m.j;
  k = pl->m.k;
  a1 = pl->m.a;  a0 = 1 - a1;
  b1 = pl->m.b;  b0 = 1 - b1;
  c1 = pl->m.c;  c0 = 1 - c1;
  v0 = VX3(i-1, j, k);
  v1 = VX3(i, j, k);
  pv->x = a0*v0 + a1*v1;
  v0 = VY3(i, j-1, k);
  v1 = VY3(i, j, k);
  pv->y = b0*v0 + b1*v1;
  v0 = VS3(i, j, k-1);
  v1 = VS3(i, j, k);
  pv->z = c0*v0 + c1*v1;
  return 0;
  }

/*==================================================================== Int3dm
*/
static int Int3dm(  /* Interpolate fields in 3 dimensions   */
LOCREC *pl,         /* location within grid                 */
MODREC *pm )        /* interpolated field                   */
  {
  float a000, a001, a010, a011, a100, a101, a110, a111;
  float a1, a0, b1, b0, c1, c0;
  int i, j, k, n;
  float *p000, *p001, *p010, *p011, *p100, *p101, *p110, *p111;
  float v1, v0, *pf;
  ARYDSC *pa;
  i = pl->m.i;
  j = pl->m.j;
  k = pl->m.k;
  if (VS3(i, j, k-1) <= -99)  return PRT_INB;   // inside body
  a1 = pl->m.a;  a0 = 1 - a1;
  b1 = pl->m.b;  b0 = 1 - b1;
  c1 = pl->m.c;  c0 = 1 - c1;
  a000 = a0*b0*c0;
  a001 = a0*b0*c1;
  a010 = a0*b1*c0;
  a011 = a0*b1*c1;
  a100 = a1*b0*c0;
  a101 = a1*b0*c1;
  a110 = a1*b1*c0;
  a111 = a1*b1*c1;
  pa = &TtDsc;
  p000 = AryPtrX(pa, i-1, j-1, k-1);
  p001 = AryPtrX(pa, i-1, j-1, k  );
  p010 = AryPtrX(pa, i-1, j  , k-1);
  p011 = AryPtrX(pa, i-1, j  , k  );
  p100 = AryPtrX(pa, i  , j-1, k-1);
  p101 = AryPtrX(pa, i  , j-1, k  );
  p110 = AryPtrX(pa, i  , j  , k-1);
  p111 = AryPtrX(pa, i  , j  , k  );
  pm->tau = a000*(*p000) + a001*(*p001) + a010*(*p010) + a011*(*p011)
          + a100*(*p100) + a101*(*p101) + a110*(*p110) + a111*(*p111);
  v0 = VX3(i-1, j, k);
  v1 = VX3(i, j, k);
  pm->wind.x = a0*v0 + a1*v1;
  v0 = VY3(i, j-1, k);
  v1 = VY3(i, j, k);
  pm->wind.y = b0*v0 + b1*v1;
  v0 = VS3(i, j, k-1);
  v1 = VS3(i, j, k);
  pm->wind.z = c0*v0 + c1*v1;
  pm->drift = WW3(i, j, k);
  pa = &PpDsc;
  p000 = AryPtrX(pa, i-1, j-1, k-1);
  p001 = AryPtrX(pa, i-1, j-1, k  );
  p010 = AryPtrX(pa, i-1, j  , k-1);
  p011 = AryPtrX(pa, i-1, j  , k  );
  p100 = AryPtrX(pa, i  , j-1, k-1);
  p101 = AryPtrX(pa, i  , j-1, k  );
  p110 = AryPtrX(pa, i  , j  , k-1);
  p111 = AryPtrX(pa, i  , j  , k  );
  pf = &pm->psi.xx;
  for (n=0; n<9; n++)
    *pf++ = a000*(*p000++) + a001*(*p001++) + a010*(*p010++) + a011*(*p011++)
          + a100*(*p100++) + a101*(*p101++) + a110*(*p110++) + a111*(*p111++);
  pa = &LlDsc;
  p000 = AryPtrX(pa, i-1, j-1, k-1);
  p001 = AryPtrX(pa, i-1, j-1, k  );
  p010 = AryPtrX(pa, i-1, j  , k-1);
  p011 = AryPtrX(pa, i-1, j  , k  );
  p100 = AryPtrX(pa, i  , j-1, k-1);
  p101 = AryPtrX(pa, i  , j-1, k  );
  p110 = AryPtrX(pa, i  , j  , k-1);
  p111 = AryPtrX(pa, i  , j  , k  );
  pf = &pm->lam.xx;
  for (n=0; n<9; n++)
    *pf++ = a000*(*p000++) + a001*(*p001++) + a010*(*p010++) + a011*(*p011++)
          + a100*(*p100++) + a101*(*p101++) + a110*(*p110++) + a111*(*p111++);
  return 0;
  }
/*================================================================= ClcModValue
*/
int ClcModValue(          /* Feld am Ort des Teilchens */
LOCREC *pl,
MODREC *pm )
  {
  int k, l;
  float c1, c0, *p1, *p0, *pi;
  if (PRTMET3) {
    k = Int3dm(pl, pm);  if (k)  return k;
    if (PrtMode & PRT_NODRIFT)  pm->drift = DefV(0,0,0);
    }
  else {
    k = pl->m.k;
    c1 = pl->m.c;  c0 = 1 - c1;
    pm->tau = c0*TT(k-1) + c1*TT(k);
    p1 = &vVV(k).x;  p0 = &vVV(k-1).x;
    pi = &pm->wind.x;
    for (l=0; l<3; l++)  *pi++ = c1*(*p1++) + c0*(*p0++);
    pm->drift = (PrtMode & PRT_NODRIFT) ? DefV(0,0,0) :  vWW(k);
    p1 = &mPP(k).xx;  p0 = &mPP(k-1).xx;
    pi = &pm->psi.xx;
    for (l=0; l<9; l++)  *pi++ = c1*(*p1++) + c0*(*p0++);
    p1 = &mLL(k).xx;  p0 = &mLL(k-1).xx;
    pi = &pm->lam.xx;
    for (l=0; l<9; l++)  *pi++ = c1*(*p1++) + c0*(*p0++);
    }
if (PrtCheck & PRT_CHKMOD) {
  fprintf(ChkFile, "MOD   tau=%1.2f  pos=(%12.3e, %12.3e, %12.3e)\n",
    pm->tau, pl->m.x, pl->m.y, pl->m.z);
  fprintf(ChkFile, "wind  : %12.5e  %12.5e  %12.5e\n",
   pm->wind.x, pm->wind.y, pm->wind.z);
  fprintf(ChkFile, "drift : %12.5e  %12.5e  %12.5e\n",
   pm->drift.x, pm->drift.y, pm->drift.z);
  fprintf(ChkFile, "psi   : %12.5e  %12.5e  %12.5e\n",
   pm->psi.xx, pm->psi.xy, pm->psi.xz);
  fprintf(ChkFile, "      : %12.5e  %12.5e  %12.5e\n",
   pm->psi.yx, pm->psi.yy, pm->psi.yz);
  fprintf(ChkFile, "      : %12.5e  %12.5e  %12.5e\n",
   pm->psi.zx, pm->psi.zy, pm->psi.zz);
  fprintf(ChkFile, "lam   : %12.5e  %12.5e  %12.5e\n",
   pm->lam.xx, pm->lam.xy, pm->lam.xz);
  fprintf(ChkFile, "      : %12.5e  %12.5e  %12.5e\n",
   pm->lam.yx, pm->lam.yy, pm->lam.yz);
  fprintf(ChkFile, "      : %12.5e  %12.5e  %12.5e\n",
   pm->lam.zx, pm->lam.zy, pm->lam.zz);
   }
  return 1;
  }

/*==================================================================== ClcLoc
*/
int ClcLoc(    /* Feststellen, in welcher Masche das Teilchen ist */
VECTOR *pp,
LOCREC *pl )
  {
  dP(ClcLoc);
  int i, j, k, rc;
  float x, y, z, s;
  if (!pp || !pl)                                       eX(0);
  rc = 0;
  x = pp->x;
  y = pp->y;
  z = pp->z;
  if (x < X0)
    if (PRTPERX)  x -= (X1-X0)*floor((x-X0)/(X1-X0));
    else  return PRT_OUT;
  if (x >= X1)
    if (PRTPERX)  x -= (X1-X0)*floor((x-X0)/(X1-X0));
    else  return PRT_OUT;
  if (y < Y0)
    if (PRTPERY)  y -= (Y1-Y0)*floor((y-Y0)/(Y1-Y0));
    else  return PRT_OUT;
  if (y >= Y1)
    if (PRTPERY)  y -= (Y1-Y0)*floor((y-Y0)/(Y1-Y0));
    else  return PRT_OUT;
  if (PRTMET3) {
    rc = GrdClcS(x, y, &z, &s);                         eG(1);
    if      (rc == GRD_SLOW)  rc = PRT_BOT;
    else if (rc == GRD_SHGH)  rc = PRT_TOP;
    }
  else {
    if      (z < 0)  { z = -z;  rc = PRT_BOT; }
    else if (ZH>0 && z>ZH) { z = ZH - (z-ZH);  rc = PRT_TOP; }
    s = z;
    }
  if (s > Zmax)  return PRT_OUT;
  pl->m.x = x;  pp->x = x;
  pl->m.y = y;  pp->y = y;
  pl->m.z = z;  pp->z = z;
  pl->m.s = s;
  for (k=1; k<=NZ; k++)  if (ZZ(k) > s)  break;
  if (k < 1)   k = 1;
  if (k > NZ)  k = NZ;
  if (NX) {
    i = 1 + (x - X0)/DD;
    if (i < 1)   i = 1;
    if (i > NX)  i = NX;
    }
  else  i = 0;
  if (NY) {
    j = 1 + (y - Y0)/DD;
    if (j < 1)   j = 1;
    if (j > NY)  j = NY;
    }
  else  j = 0;
  pl->m.i = i;
  pl->m.j = j;
  pl->m.k = k;
  pl->m.a = (NX) ? (x - X0)/DD - i + 1 : 0;
  pl->m.b = (NY) ? (y - Y0)/DD - j + 1 : 0;
  pl->m.c = (s - ZZ(k-1))/(ZZ(k) - ZZ(k-1));
  if (PRTMET3 && (k))
      if (VS3(i,j,k-1) <= -99)  rc = PRT_INB;
  if (x < A0)  i = 0;
  else {
    i = 1 + (x - A0)/DA;
    if (i > MX)  i = 0;
    }
  if (y < B0)  j = 0;
  else {
    j = 1 + (y - B0)/DA;
    if (j > MY)  j = 0;
    }
  if (s < C0)  k = 0;
  else {
    k = 1 + (s - C0)/DC;
    if (k > MZ)  k = 0;
    }
  pl->s.i = i;
  pl->s.j = j;
  for (k=0; k<=MZ; k++)  if (CC(k) > s)  break;
  if (k > MZ)  k = 0;
  pl->s.k = k;
if (PrtCheck & PRT_CHKLOC) fprintf(ChkFile,
 "LOC (%3.1f,%3.1f,%3.1f,%3.1f) in (%d,%d,%d) & (%1.2f,%1.2f,%1.2f)\n",
 pl->m.x, pl->m.y, pl->m.z, pl->m.s, pl->m.i, pl->m.j, pl->m.k,
 pl->m.a, pl->m.b, pl->m.c);
  return rc;
eX_0:
  eMSG("invalid parameter! (%s)\n", ePGMs);
eX_1:
  eMSG("can't locate particle! (%s)\n", ePGMs);
  }

/*====================================================================== NewPtl
*/
PTLREC *NewPtl( void )            /* Neues Teilchen zur Verfgung stellen */
  {
  dP(NewPtl);
  PTLREC *pp;
  pp = ALLOC(sizeof(PTLREC));  if (!pp)         eX(1);
  pp->next = PtlTab;
  if (PtlTab)  PtlTab->prev = pp;
  PtlTab = pp;
  Nptl++;
  return pp;
eX_1:
  vMsg("can't allocate particle (Nptl=%d)! (%s)\n", Nptl, ePGMs);
  return NULL;
  }

/*===================================================================== InitPtl
*/
static void prnmat(MATRIX m)
  {
  fprintf(ChkFile, "%12.5e  %12.5e  %12.5e\n", m.xx, m.xy, m.xz);
  fprintf(ChkFile, "%12.5e  %12.5e  %12.5e\n", m.yx, m.yy, m.yz);
  fprintf(ChkFile, "%12.5e  %12.5e  %12.5e\n", m.zx, m.zy, m.zz);
  }

int InitPtl( PTLREC *pp )       /* Teilchen-Geschwindigkeit initialisieren */
  {
  dP(InitPtl);
  LOCREC l;
  float *p1, *p0, *pi;
  MATRIX eps;
  VECTOR rvc;
  ARYDSC *pa;
  float a000, a001, a010, a011, a100, a101, a110, a111;
  float a1, a0, b1, b0, c1, c0;
  int i, j, k, n, flag;
  float *p000, *p001, *p010, *p011, *p100, *p101, *p110, *p111;

  if (pp->f == PRT_OUT)  return PRT_OUT;
  flag = ClcLoc(&(pp->r), &l);                              eG(1);
  if (flag) {
    pp->f = PRT_OUT;
    return PRT_OUT;
    }
  rvc = DefV(RndNrm(), RndNrm(), RndNrm());
  pi = &eps.xx;
  if (!PRTMET3) {
    k  = l.m.k;
    c1 = l.m.c;
    c0 = 1 - c1;
    p1 = &mEE(k).xx;  p0 = &mEE(k-1).xx;
    for (n=0; n<9; n++)  *pi++ = c1*(*p1++) + c0*(*p0++);
    pp->v = DotMV(eps, rvc);
    if (PrtCheck & PRT_CHKINI)  goto print_out;
    return 0;
    }
  i = l.m.i;
  j = l.m.j;
  k = l.m.k;
  a1 = l.m.a;  a0 = 1 - a1;
  b1 = l.m.b;  b0 = 1 - b1;
  c1 = l.m.c;  c0 = 1 - c1;
  a000 = a0*b0*c0;
  a001 = a0*b0*c1;
  a010 = a0*b1*c0;
  a011 = a0*b1*c1;
  a100 = a1*b0*c0;
  a101 = a1*b0*c1;
  a110 = a1*b1*c0;
  a111 = a1*b1*c1;
  pa = &EeDsc;
  p000 = AryPtrX(pa, i-1, j-1, k-1);
  p001 = AryPtrX(pa, i-1, j-1, k  );
  p010 = AryPtrX(pa, i-1, j  , k-1);
  p011 = AryPtrX(pa, i-1, j  , k  );
  p100 = AryPtrX(pa, i  , j-1, k-1);
  p101 = AryPtrX(pa, i  , j-1, k  );
  p110 = AryPtrX(pa, i  , j  , k-1);
  p111 = AryPtrX(pa, i  , j  , k  );
  for (n=0; n<9; n++)
    *pi++ = a000*(*p000++) + a001*(*p001++) + a010*(*p010++) + a011*(*p011++)
          + a100*(*p100++) + a101*(*p101++) + a110*(*p110++) + a111*(*p111++);
  pp->v = DotMV(eps, rvc);
  if (PrtCheck & PRT_CHKINI)  goto print_out;
  return 0;
print_out:
  fprintf(ChkFile, "INI r=(%8.5f,%8.5f,%8.5f) v=(%12.5e,%12.5e,%12.5e)\n",
    rvc.x, rvc.y, rvc.z, pp->v.x, pp->v.y, pp->v.z);
  return 0;
eX_1:
  eMSG("can't initialize particle! (%s)\n", ePGMs);
  }

/*====================================================================== GetPtl
*/
int GetPtl( void )
  {
  dP(GetPtl);
  static int nn;
  int n, np, nout;
  static float rad2deg = 57.295780;
  float da, db, dz, p, x, y, z;
  float sip, cop, gp, dt;
  PTLREC *pp;
  if (EQ <= 0)  return 0;
  if (PQ) {
    p = PQ/rad2deg;
    sip = sin(p);
    cop = cos(p);
    }
  else {
    sip = 0;
    cop = 1;
    }
  np = 0.5 + DT*RP;
  if (np < 1)  return 0;
  dt = DT/np;
  gp = EQ*DT/np;
  nout = 0;
  for (n=0; n<np; n++) {
    pp = NewPtl();  if (!pp)                            eX(1);
    pp->f = 0;
    if (PrtCheck & PRT_CHKREG) {
      int nps;
      nps = sqrt(np + 0.1);
      da = (0.5 + n%nps)*AQ/nps;
      db = (0.5 + n/nps)*BQ/nps;
      }
    else {
      da = (AQ) ? AQ*RndEqu01() : 0;
      db = (BQ) ? BQ*RndEqu01() : 0;
      }
    dz = (CQ) ? CQ*RndEqu01() : 0;
    x = XQ + cop*da - sip*db;
    y = YQ + sip*da + cop*db;
    z = HQ + dz + SQ*da;
    if (PRTMET3) {
      if (x>=X0 && x<=X1 && y>=Y0 && y<=Y1)  z += GrdBxy(x, y);
      else { pp->f = PRT_OUT;  nout++; }
      }
    pp->r.x = x;
    pp->r.y = y;
    pp->r.z = z;
    pp->v.x = 0;
    pp->v.y = 0;
    pp->v.z = -999;
    pp->g0 = gp;
    pp->t0 = Tm + ((PrtCheck & PRT_CHKREG) ? 0.5*DT : (n+0.5)*dt);
    pp->g = pp->g0;
    pp->t = pp->t0;
    pp->u.x = UU(0)*(1 + UD*RndNrm());
    pp->u.y = UU(1)*(1 + UD*RndNrm());
    pp->u.z = UU(2)*(1 + UD*RndNrm());
    pp->n = ++nn;
    }
  if (nout)  vMsg("GET: %d particles outside!\n", nout);
  return np;
eX_1:
  eMSG("*** Neues Teilchen kann nicht erzeugt werden! (%s)\n", ePGMs);
  }

/*=================================================================== FirstPtl
*/
PTLREC *FirstPtl( void )          /* Pointer auf erstes Teilchen */
  {
  return PtlTab;
  }

/*==================================================================== NextPtl
*/
PTLREC *NextPtl( PTLREC *pp )    /* Pointer auf nchstes Teilchen */
  {
  return pp->next;
  }

/*=================================================================== ClearPtl
*/
PTLREC *ClearPtl( PTLREC *pp )    /* Pointer auf nchstes Teilchen */
  {
  PTLREC *pn;
  pn = pp->next;
  if (pp->next)  (pp->next)->prev = pp->prev;
  if (pp->prev)  (pp->prev)->next = pp->next;
  else  PtlTab = pp->next;
  FREE(pp);
  Nptl--;
  return pn;
  }

/*================================================================ ClearAllPtl
*/
int ClearAllPtl( void )
  {
  PTLREC *pp, *pn;
  int n;
  n = 0;
  pp = PtlTab;
  while (pp) {
    pn = pp->next;
    FREE(pp);
    n++;
    pp = pn;
    }
  Nptl = 0;
  PtlTab = NULL;
  return n;
  }

/*===================================================================== PrnPtl
*/
int PrnPtl( void )
  {
  int n;
  PTLREC *pp;
  pp = FirstPtl();
  for (n=1; (pp); n++) {
    fprintf(ChkFile,
     "%4d %6d : %6.1f %6.1f %6.1f %8.1f  %6.2f %6.2f %6.2f \n",
      n, pp->n, pp->r.x, pp->r.y, pp->r.z, pp->t, pp->v.x, pp->v.y, pp->v.z );
    pp = NextPtl(pp);
    }
  return n-1;
  }

/*==================================================================== AddDose
*/
static int AddDose(             /* Beitrag zur Dosis aufaddieren        */
PTLREC *pa, LOCREC *pold,       /* Alte Position                        */
PTLREC *pe, LOCREC *pnew,       /* Neue Position                        */
int flag )                      /* Kennzeichnung des Schrittes          */
  {
  int i, j, k;
  float d;
  d = 0.5*(pe->t - pa->t);
  i = pold->s.i;
  j = pold->s.j;
  k = pold->s.k;
  if (i && j && k)  CON(i,j,k) += d*pa->g;
  i = pnew->s.i;
  j = pnew->s.j;
  k = pnew->s.k;
  if (i && j && k)  CON(i,j,k) += d*pe->g;
  return 1;
  }

/*===================================================================== AddIcp
*/
static int AddIcp(              /* Beitrag zur Interzeption aufaddieren */
PTLREC *pa, LOCREC *pold,       /* Alte Position                        */
PTLREC *pe, LOCREC *pnew,       /* Neue Position                        */
VECTOR ss )                     /* Neue Position durch Advektion        */
  {
  int i, j, k;
  float dd, li, dg, dx, dy, dz;
  i = pold->m.i;
  j = pold->m.j;
  k = pold->m.k;
  if (i==0 || j==0 || k==0)  return 0;
  li = LI3(i,j,k);
  if (li <= 0)  return 0;
  i = pold->s.i;
  j = pold->s.j;
  k = pold->s.k;
  if (i==0 || j==0 || k==0)  return 0;
  dx = ss.x - pa->r.x;
  dy = ss.y - pa->r.y;
  dz = ss.z - pa->r.z;
  dd = sqrt(dx*dx + dy*dy + dz*dz);
  dg = (dd > 0.2*li) ? 1-exp(-dd/li) : dd/li;
  dg *= pa->g;
  if (dg > pe->g)  dg = pe->g;
  ICP(i,j,k) += dg;
  pe->g -= dg;
  return 1;
  }

/*================================================================= AddDryDepo
*/
static int AddDryDepo(          /* Beitrag zur trockenen Deposition */
PTLREC *pa, LOCREC *pold,       /* Alte Position                    */
PTLREC *pe, LOCREC *pnew )      /* Neue Position                    */
  {
  int i, j;
  float dg, pd;
  pd = PD;
  if (PdDsc.start) {
    i = pold->m.i;
    j = pold->m.j;
    if (i && j)  pd = PD2(i,j);
    }
  dg = pd*pa->g;
  if (dg > pe->g)  dg = pe->g;
  i = pold->s.i;
  j = pold->s.j;
  if (i && j)  DRY(i,j) += 0.5*dg;
  i = pnew->s.i;
  j = pnew->s.j;
  if (i && j)  DRY(i,j) += 0.5*dg;
  pe->g -= dg;
  return 1;
  }

/*================================================================= AddWetDepo
*/
static int AddWetDepo(          /* Beitrag zur Nassen Deposition        */
PTLREC *pa, LOCREC *pold,       /* Alte Position                        */
PTLREC *pe, LOCREC *pnew )      /* Neue Position                        */
  {
  int i, j;
  float d, dg, da, de;
  dg = 0;
  d = 0.5*(pe->t - pa->t)*RW;
  da = d*pa->g;
  de = d*pe->g;
  dg = da + de;
  if (dg > pe->g) {
    da *= pe->g/dg;
    de *= pe->g/dg;
    dg = pe->g;
    }
  i = pold->s.i;
  j = pold->s.j;
  if (i && j)  WET(i,j) += d*pa->g;
  i = pnew->s.i;
  j = pnew->s.j;
  if (i && j)  WET(i,j) += d*pe->g;
  pe->g -= dg;
  return 1;
  }

/*=================================================================== PosAdvec
*/
int PosAdvec( GRDLOCREC *pl, MODREC *pm, VECTOR *pv )
  {
  dP(PosAdvec);
  float tau, x, y, z, s;
  int flag;
  LOCREC l2;
  VECTOR vm, v2;
  tau = pm->tau;
  x = pl->x + tau*(pm->wind).x;
  y = pl->y + tau*(pm->wind).y;
  if (x < X0)
    if (PRTPERX)  x -= (X1-X0)*floor((x-X0)/(X1-X0));
    else  return PRT_OUT;
  if (x >= X1)
    if (PRTPERX)  x -= (X1-X0)*floor((x-X0)/(X1-X0));
    else  return PRT_OUT;
  if (y < Y0)
    if (PRTPERY)  y -= (Y1-Y0)*floor((y-Y0)/(Y1-Y0));
    else  return PRT_OUT;
  if (y >= Y1)
    if (PRTPERY)  y -= (Y1-Y0)*floor((y-Y0)/(Y1-Y0));
    else  return PRT_OUT;
  pv->x = x;
  pv->y = y;
  if (!PRTMET3) {
      pv->z = pl->s + tau*(pm->wind).z;
      return 0;
      }
  if (PRTUSE_VV) {
    flag = GrdZ(x, y, pl->s, &z);           eG(1);
    z += tau*(pm->wind).z;
    }
  else {
    s = pl->s + tau*(pm->wind).z;
    flag = GrdZ(x, y, s, &z);               eG(1);
    }
  if (flag & GRD_OUT)  return PRT_OUT;
  pv->z = z;
  if (!PRTVELCOR)  return 0;

  /* higher order correction of the particle path */

  flag = ClcLoc(pv,&l2);  if (flag)  return flag;
  Int3dv(&l2, &v2);
  vm = DefV(0.5*((pm->wind).x+v2.x),
            0.5*((pm->wind).y+v2.y),
            0.5*((pm->wind).z+v2.z));
  x = pl->x + tau*vm.x;
  y = pl->y + tau*vm.y;
  if (x < X0)
    if (PRTPERX)  x -= (X1-X0)*floor((x-X0)/(X1-X0));
    else  return PRT_OUT;
  if (x >= X1)
    if (PRTPERX)  x -= (X1-X0)*floor((x-X0)/(X1-X0));
    else  return PRT_OUT;
  if (y < Y0)
    if (PRTPERY)  y -= (Y1-Y0)*floor((y-Y0)/(Y1-Y0));
    else  return PRT_OUT;
  if (y >= Y1)
    if (PRTPERY)  y -= (Y1-Y0)*floor((y-Y0)/(Y1-Y0));
    else  return PRT_OUT;
  pv->x = x;
  pv->y = y;
  if (PRTUSE_VV) {
    flag = GrdZ(x, y, pl->s, &z);           eG(2);
    z += tau*vm.z;
    }
  else {
    s = pl->s + tau*vm.z;
    flag = GrdZ(x, y, s, &z);               eG(2);
    }
  if (flag & GRD_OUT)  return PRT_OUT;
  pv->z = z;
  return 0;
eX_1:  eX_2:
  eMSG("can't locate! (%s)\n", ePGMs);
  }

/*================================================================== NewUbias
*/
VECTOR NewUbias(VECTOR ob, float tau)
  {
  VECTOR nb;
  if (tau < UT)  nb.x = ob.x*(1-tau/UT);  else nb.x = 0;
  if (tau < UT)  nb.y = ob.y*(1-tau/UT);  else nb.y = 0;
  if (tau < UT)  nb.z = ob.z*(1-tau/UT);  else nb.z = 0;
  return nb;
  }

/*===================================================================== RunPtl
*/
#define CHECK(a)  if (PrtCheck & (a)) fprintf(ChkFile, \
    "PTL %4d (%d,%04x): T=%5.1f  R=(%5.2f,%5.2f,%5.2f) X=(%5.1f,%5.1f,%5.1f)  V=(%5.2f,%5.2f,%5.2f)\n", \
     ip, trial, e.f, e.t, rr.x, rr.y, rr.z, e.r.x, e.r.y, e.r.z, e.v.x, e.v.y, e.v.z )

#define TRACE(f)  if (PrtCheck & (f)) fprintf(ChkFile, \
    "PTL %6d %8d : %10.2f %10.2f %10.2f %10.2f\n", ip, a.n, a.r.x, a.r.y, a.r.z, a.t);

int RunPtl( void )   /* Serie von Zeitschritten durchfhren */
  {
  dP(RunPtl);
  int trial, ip, nout, dry_depo, wet_depo, interception, first_step;
  double t2;
  VECTOR rr, vv, ss;
  PTLREC *pp, a, e;
  LOCREC lnew, lold;
  MODREC m;
  clock_t first, next, now;
  int ninit = 0;
  if (Ticks > 0) {
    first = clock();
    next = first + Ticks*CLOCKS_PER_SEC;
    }
  dry_depo = (PD>0 || (PdDsc.start));       // Mit Deposition rechnen?
  wet_depo = (RW > 0);
  interception = (NULL != LiDsc.start);
  memset(&lold, 0, sizeof(lold));           // Strukturen initialisieren ...
  memset(&lnew, 0, sizeof(lnew));
  memset(&m, 0, sizeof(m));
  nout = 0;
  t2 = Tm + DT;                             // Endzeitpunkt festlegen
  ip = 0;                                   // Ptl.-Zhler initialisieren
  pp = FirstPtl();                          // Erstes Partikel holen
  while (pp) {                              // Alle Partikel durchgehen
    ip++;                                   // Partikel mitzhlen
    if (pp->v.z <= -999) {                  // Neues Partikel
      first_step = 1;                       // Merken
      //pp->rnd = RndSeed();
      InitPtl(pp);              eG(1);      // und initialisieren
      ninit++;                              // Mitzhlen
      }
    else  first_step = 0;                   // Kein neues Partikel
    e = *pp;                                // Position als Endpos. merken
    if (e.f & (PRT_OUT|PRT_INB))  goto done;// Falls auerhalb: fertig
    e.f = ClcLoc(&e.r, &lold);  eG(2);      // Position im Raster berechnen
    if (e.f & (PRT_OUT|PRT_INB))  goto done;// Falls auerhalb: fertig
    a = e;                                  // Startposition definieren
    //RndSet(a.rnd);
    while (a.t <= t2) {                     // Zeitschritte durchfhren
      TRACE(PRT_CHKTRC);
      rr = DefV(RndNrm(),RndNrm(),RndNrm());      // Zufallszahlen ziehen
      ClcModValue(&lold, &m);               // Felder am Ort des Partikels
      if (first_step)                       // Falls erster Zeitschritt,
        m.tau *= (0.5+RndEqu01());          // statistisch variieren
      first_step = 0;                       // In Zukunft nicht mehr
      e.f = PosAdvec(&(lold.m), &m, &ss);  eG(5);   // Nur Advektion
      if (e.f & PRT_OUT)  break;
      vv = DotMV(m.psi, e.v);               // Neues v = Anteil altes v
      vv = AddVV(vv, m.drift);              // + Drift-Geschwindigkeit
      vv = AddVV(vv, DotMV(m.lam,rr));      // + zuflliger Anteil
      e.v = vv;                             // Neues v
      vv = AddVV(vv, e.u);                  // Aufgeprgte Geschwindigkeit
      vv.z -= VS;                           // Sedimentation
      e.r = AddVV(ss, MulSV(m.tau,vv));     // Neue Position
      e.t = a.t + m.tau;                    // Zeit weitersetzen
      e.g = a.g;                            // Zunchst Masse unverndert
      e.f = ClcLoc(&e.r, &lnew); eG(3);     // Neue Position im Raster
      trial = 0;
      while (e.f & PRT_INB) {               // Falls innerhalb eines Krpers:
        trial++;                            // Versuche mitzhlen
        CHECK(PRT_CHKTRL);
        if (trial > 100) {                  // Der Fall ist hoffnungslos:
          vMsg("trial>100   (ip=%d)\n",ip); // melden,
          CHECK(PRT_CHKTRL);
          e.f |= PRT_OUT;                   // fr ungltig erklren ...
          break;                            // und nchstes nehmen.
          }
        InitPtl(&a);         eG(4);         // Neu initialisieren
        e = a;
        e.r = AddVV(a.r, MulSV(m.tau,a.v)); // Neue Position
        e.t = a.t + m.tau;                  // Zeit weitersetzen
        e.f = ClcLoc(&e.r, &lnew); eG(3);   // Neue Position im Raster
        }
      CHECK(PRT_CHKPTL);
      if (e.f & PRT_OUT)  break;            // Falls aussortiert, abbrechen
      e.u = NewUbias(e.u, m.tau);           // Neue Werte fr U berechnen
      if (e.f & PRT_BOT) {                  // Erdboden wurde getroffen
        if (dry_depo)                       // Trockene Deposition ...
          AddDryDepo(&a,&lold, &e,&lnew);   // aufadddieren.
        e.v.z = -e.v.z;                     // elastisch reflektieren
        }
      else if (e.f & PRT_TOP) {             // Obergrenze wurde getroffen
        e.v.z = -e.v.z;                     // Elastische Reflexion
        }
      if (interception)                     // Falls Interzeption:
        AddIcp(&a,&lold, &e,&lnew, ss);     // Beitrag aufaddieren
      AddDose(&a,&lold, &e,&lnew, e.f);     // Dosis-Beitrag aufaddieren
      if (wet_depo)                         // Nasse Deposition aufaddieren
        AddWetDepo(&a,&lold, &e,&lnew);
      if (e.g < QP*e.g0) {                  // Partikel hat zu viel Masse ..
        e.f |= PRT_OUT;                     // .. verloren, aussortieren!
        break;                              // Nchstes Partikel nehmen
        }
      a = e;                                // Neuer Startpunkt
      a.f = 0;
      lold = lnew;
      trial = 0;
      }
done:
    //a.rnd = RndSeed();
    *pp = e;                            // Endposition merken
    if (pp->f & PRT_OUT) {              // Wenn Rechengebiet verlassen,
      CHECK(PRT_CHKOUT);
      pp = ClearPtl(pp);                // Partikel lschen
      nout++;                           // Anzahl merken
      }
    else  pp = NextPtl(pp);             // Nchstes Partikel bearbeiten
    if (Ticks > 0) {                    // Falls Kontrollausgabe:
      now = clock();                    // Uhrzeit feststellen
      if (now > next) {                 // Falls Kontrollzeit erreicht:
        printf("t=%s  ip=%d\r", MsgDateString(Tm/(24.0*3600)), ip);
        fflush(stdout);
        next += Ticks*CLOCKS_PER_SEC;   // Nchster Kontrollzeitpunkt
        }
      }
    }
  if ((ninit>0) && (MsgVerbose&0x10))
    vMsg("%5d particles initialized\n", ninit);
  if (Ticks > 0)  printf("t=%s  ip=%d\n", MsgDateString(Tm/(24.0*3600)), ip);
  if ((nout) && (MsgVerbose&0x02))  vMsg("%5d particles removed\n", nout);
  return ip;
eX_1: eX_2: eX_3: eX_4: eX_5:
  eMSG("can't move particle! (%s)\n", ePGMs);
  }
#undef CHECK

/*===================================================================== ClcCon
*/
int ClcCon( char *arg )            /* Berechnung der Konzentration */
  {
  dP(ClcCon);
  int n;
  n = GetPtl();                                         eG(3);
  if (MsgVerbose & 0x02)  vMsg("%5d particles created\n", n);
  if (PrtCheck & PRT_CHKLST)  PrnPtl();
  if (DT <= 0)  return 0;
  n = RunPtl();                                         eG(4);
  if (MsgVerbose & 0x02)  vMsg("%5d particles processed\n", n);
  Tm += DT;
  T2 = Tm/(24.0*3600);
  return 0;
eX_3:
  eMSG("no particles generated! (%s)\n", ePGMs);
eX_4:
  eMSG("error in particle driver! (%s)\n", ePGMs);
  }

/*===================================================================== NrmDos
*/
static float SmplHeight( int i, int j )
  {
  dP(SmplHeight);
  float z00, z01, z10, z11, zmm, z;
  float x0, x1, y0, y1, xm, ym;
  x0 = A0 + (i-1)*DA;
  x1 = A0 + i*DA;
  y0 = B0 + (j-1)*DA;
  y1 = B0 + j*DA;
  xm = 0.5*(x0 + x1);
  ym = 0.5*(y0 + y1);
  z00 = GrdBxy(x0, y0);                 eG(1);
  z01 = GrdBxy(x0, y1);                 eG(2);
  z10 = GrdBxy(x1, y0);                 eG(3);
  z11 = GrdBxy(x1, y1);                 eG(4);
  zmm = GrdBxy(xm, ym);                 eG(5);
  z = (z00 + z01 + z10 + z11 + 2*zmm)/6;
  return z;
eX_1: eX_2: eX_3: eX_4: eX_5:
  eMSG("can't compute average height for (%d,%d)! (%s)\n", i, j, ePGMs);
  }

ARYDSC *NrmDos( ARYDSC *pa )
  {
  dP(NrmDos);
  int nx, ny, nz, i, j, k, tr3;
  double a, *pf, secs, z;  /*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
  nx = pa->bound[0].hgh;
  ny = pa->bound[1].hgh;
  secs = (T2-T1)*24*3600;
  if (pa->numdm == 2) {
    a = DA*DA*secs;
    if (a > 0)  a = 1/a;
    AryCreate(&TmpDsc, sizeof(float), 2, 1, nx, 1, ny);         eG(1);
    for (i=1; i<=nx; i++)
      for (j=1; j<=ny; j++) {
        pf = AryPtr(pa, i, j);  if (!pf)                        eX(2);
        TMP2(i,j) = a*(*pf);
        }
    return &TmpDsc;
    }
  if (pa->numdm == 3) {
    tr3 = (PRTMET3) && (ZH > 0);
    nz = pa->bound[2].hgh;
    AryCreate(&TmpDsc, sizeof(float), 3, 1, nx, 1, ny, 1, nz);  eG(3);
    for (i=1; i<=nx; i++)
      for (j=1; j<=ny; j++) {
        if (tr3) {
          z = SmplHeight(i, j);                                 eG(5);
          }
        for (k=1; k<=nz; k++) {
          a = secs*DA*DA*(CC(k)-CC(k-1));
          if (tr3)  a *= (ZH - z)/SH;
          if (a > 0)  a = 1/a;
          pf = AryPtr(pa, i, j, k);  if (!pf)                   eX(4);
          TMP3(i,j,k) = a*(*pf);
          }
        }
    return &TmpDsc;
    }
  return NULL;
eX_1:
  vMsg("can't create tmp(%d,%d)! (%s)\n", nx, ny, ePGMs);
  return NULL;
eX_2:
  vMsg("index (%d,%d) out of range! (%s)\n", i, j, ePGMs);
  return NULL;
eX_3:
  vMsg("can't create tmp(%d,%d,%d)! (%s)\n", nx, ny, nz, ePGMs);
  return NULL;
eX_4: eX_5:
  vMsg("index (%d,%d,%d) out of range! (%s)\n", i, j, k, ePGMs);
  return NULL;
  }

/*================================================================== WriteData
*/
static char *trmfixed( double x )
  {
  static char s[80];
  int i;
  if (x>1.e6 || x<-1.e6)  sprintf(s, "%10.3le", x);
  else {
    sprintf(s, "%1.2lf", x);
    for (i=strlen(s)-1; i>0; i--) {
      if (s[i] == '0')  s[i] = 0;
      else {
        if (s[i] == '.')  s[i] = 0;
        break;
        }
      }
    }
  return s;
  }

int WriteData(          /* Daten ausschreiben   */
char *arg )
  {
  dP(WriteData);
  ARYDSC *pa;
  char fname[256], s[80], *tt, *pc, *buf;
  float xa, ya, dd, *zz, d;
  int n, i, i0, i1, j, j0, j1, k, k0, k1, z_only;
  TXTSTR usrhdr = { NULL, 0 };

  MsgLow(arg);
  pc = strtok(arg, " +\t\r\n");
  while (pc) {
    if      (!strcmp(pc, "cnc")) {
      pa = NrmDos(&ConDsc);
      tt = "concentration (g/m3)"; }
    else if (!strcmp(pc, "icp")) {
      pa = NrmDos(&IcpDsc);
      tt = "interception (g/s*m3)"; }
    else if (!strcmp(pc, "dry")) {
      pa = NrmDos(&DryDsc);
      tt = "dry deposition (g/s*m2)"; }
    else if (!strcmp(pc, "wet")) {
      pa = NrmDos(&WetDsc);
      tt = "wet deposition (g/s*m2)"; }
    else if (!strcmp(pc, "sgm")) {
      pa = &SsDsc;
      tt = "tensor sigma (m2/s2)"; }
    else if (!strcmp(pc, "lmb")) {
      pa = &LlDsc;
      tt = "tensor lambda (m/s)"; }
    else if (!strcmp(pc, "wnd")) {
      pa = &VvDsc;
      tt = "vector wind velocity (m/s)"; }
    else if (!strcmp(pc, "drf")) {
      pa = &WwDsc;
      tt = "vector drift velocity (m/s)"; }
    else if (!strcmp(pc, "kpp")) {
      pa = &KkDsc;
      tt = "tensor kappa (diffusion coefficient) (m2/s)"; }
    else if (!strcmp(pc, "psi")) {
      pa = &PpDsc;
      tt = "tensor psi (1)"; }
    else if (!strcmp(pc, "eps")) {
      pa = &EeDsc;
      tt = "tensor epsilon (m/s)"; }
    else if (!strcmp(pc, "bm")) {
      pa = &(pGrd->bmdsc);
      tt = "centered surface grid (m)"; }
    else if (!strcmp(pc, "pdp")) {
      pa = &PdDsc;
      tt = "deposition probability (1)"; }
    else if (!strcmp(pc, "vs")) {
      pa = &VsDsc;
      tt = "vertical velocity component (m/s)"; }
    else if (!strcmp(pc, "tau")) {
      pa = &TtDsc;
      tt = "time step (s)"; }
    else                                                  eX(0);
    if (!pa)                                              eX(1);
    z_only = 0;
    if (strstr("cnc+dry+wet", pc)) {
      xa = A0;
      ya = B0;
      dd = DA;
      zz = &CC(0); }
    else {
      z_only = (0 == PRTMET3);
      xa = X0;
      ya = Y0;
      dd = DD;
      zz = &ZZ(0); }

    if (!strncmp(FI, "LOG", 3))  sprintf(fname, FI, WC+AddWrite);
    else {
        if (WC > 0)  sprintf(s, "%s%04d", pc, WC+AddWrite);
        else         strcpy(s, (*FI) ? FI : pc);
        strcpy(fname, MakeFileName(s, ""));
    }
    n = pa->numdm;  if (n < 1)                              eX(3);
    TxtClr(&usrhdr);
    if (*ID) {
        sprintf(s, "id\t%s\n", ID);     TxtCat(&usrhdr, s);
    }
    if (*MO) {
        sprintf(s, "mode\t%s\n", MO);   TxtCat(&usrhdr, s);
    }
    if (*SE) {
        sprintf(s, "sequ\t%s\n", SE);   TxtCat(&usrhdr, s);
    }
    if (*FO) {
        sprintf(s, "form\t%s\n", FO);   TxtCat(&usrhdr, s);
    }
    if (FA > 0) {
        sprintf(s, "fact\t%10.3e\n", FA);   TxtCat(&usrhdr, s);
    }
    sprintf(s, "vldf V\n");                 TxtCat(&usrhdr, s);
    sprintf(s, "xa %s\n", trmfixed(xa));    TxtCat(&usrhdr, s);
    sprintf(s, "ya %s\n", trmfixed(ya));    TxtCat(&usrhdr, s);
    sprintf(s, "dd %s\n", trmfixed(dd));    TxtCat(&usrhdr, s);
    sprintf(s, "it \"%s\"\n", tt);          TxtCat(&usrhdr, s);
    sprintf(s, "t1 \"%s\"\n", MsgDateString(T1));   TxtCat(&usrhdr, s);
    sprintf(s, "t2 \"%s\"\n", MsgDateString(T2));   TxtCat(&usrhdr, s);
    buf = TxtAlloc(4000);
    if (n>0 && !z_only) {
      i0 = pa->bound[0].low;
      i1 = pa->bound[0].hgh;
      d = (i0 > 0) ? -0.5 : 0;
      strcpy(buf, "xl");
      for (i=i0; i<=i1; i++) {
        strcpy(s, "\t");
        strcat(s, trmfixed(xa+(i+d)*dd));
        if (strlen(buf)+strlen(s) < 3998)  strcat(buf, s);
        }
      strcat(buf, "\n");
      TxtCat(&usrhdr, buf);
      }
    if (n>1 && !z_only) {
      j0 = pa->bound[1].low;
      j1 = pa->bound[1].hgh;
      d = (j0 > 0) ? -0.5 : 0;
      strcpy(buf, "yl");
      for (j=j0; j<=j1; j++) {
        strcpy(s, "\t");
        strcat(s, trmfixed(ya+(j+d)*dd));
        if (strlen(buf)+strlen(s) < 3998)  strcat(buf, s);
        }
      strcat(buf, "\n");
      TxtCat(&usrhdr, buf);
      }
    if (n>2 || z_only) {
      k0 = pa->bound[(z_only) ? 0 : 2].low;
      k1 = pa->bound[(z_only) ? 0 : 2].hgh;
      strcpy(buf, "zl");
      for (k=k0; k<=k1; k++) {
        strcpy(s, "\t");
        strcat(s, trmfixed((k0>0) ? 0.5*(zz[k-1]+zz[k]) : zz[k]));
        if (strlen(buf)+strlen(s) < 3998)  strcat(buf, s);
        }
      strcat(buf, "\n");
      TxtCat(&usrhdr, buf);
      }
    FREE(buf);
    DmnWrite(fname, usrhdr.s, pa);                          eG(4);
    if (pa == &TmpDsc)  AryFree(pa);
    if (strncmp(FI, "LOG", 3))  *FI = 0;
    *ID = 0;
    pc = strtok(NULL, " +\t\r\n");
    }
  return 0;
eX_0:
  eMSG("unknown array name %s! (%s)\n", pc, ePGMs);
eX_1:
  eMSG("can't normalize %s! (%s)\n", pc, ePGMs);
eX_3:
  eMSG("invalid number of dimensions (=%d) in array %s! (%s)\n",
    pa->numdm, pc, ePGMs);
eX_4:
  eMSG("can't write array %s! (%s)\n", pc, ePGMs);
  }

/*==================================================================== ReadData
*/
static void ShowArray( char *n, ARYDSC *pa )
  {
  float *pf;
  int i, j, k;
  char s[80];
  printf("array %s\n", n);
  while (1) {
    fputs("i j k :", stdout);  gets(s);
    if (!*s)  break;
    sscanf(s, "%d%d%d", &i, &j, &k);
    pf = AryPtr(pa, i, j, k);
    if (!pf)  printf("out of range!\n");
    else printf("%s(%d,%d,%d) = %12.3e\n", n, i, j, k, *pf);
    }
  }


int ReadData(          /* Daten einlesen   */
char *arg )
  {
  dP(ReadData);
  ARYDSC *pa;
  char fname[256], s[80], *tt, *pc, *pn, nn[80];
  TXTSTR uhdr = { NULL, 0 };
  TXTSTR shdr = { NULL, 0 };
  int l;

  pn = arg;
  pc = MsgTok(pn, &pn, " +\t\r\n", &l);
  while (pc) {
    *nn = 0;
    strncat(nn, pc, l);
    if      (!strcmp(nn, "bp")) {
      pa = &(pGrd->bpdsc);
      tt = "surface elevation at gridpoints (m)"; }
    else if (!strcmp(nn, "bm")) {
      pa = &(pGrd->bmdsc);
      tt = "surface elevation at midpoints (m)"; }
    else if (!strcmp(nn, "vx")) {
      AryFree(&VsDsc);
      pa = &VxDsc;
      tt = "wind velocity in x-direction (m/s)"; }
    else if (!strcmp(nn, "vy")) {
      AryFree(&VsDsc);
      pa = &VyDsc;
      tt = "wind velocity in y-direction (m/s)"; }
    else if (!strcmp(nn, "vs")) {
      pa = &VsDsc;
      tt = "wind velocity in s-direction (m/s)"; }
    else if (!strcmp(nn, "vv")) {
      pa = &VsDsc;
      PrtMode |= PRT_USE_VV;
      tt = "wind velocity in z-direction (m/s)"; }
    else if (!strcmp(nn, "su")) {
      pa = &SuDsc;
      tt = "wind fluctuation in u-direction (m/s)"; }
    else if (!strcmp(nn, "sv")) {
      pa = &SvDsc;
      tt = "wind fluctuation in v-direction (m/s)"; }
    else if (!strcmp(nn, "sw")) {
      pa = &SwDsc;
      tt = "wind fluctuation in w-direction (m/s)"; }
    else if (!strcmp(nn, "sc")) {
      pa = &ScDsc;
      tt = "wind fluctuation sqrt(-<u'w'>) (m/s)"; }
    else if (!strcmp(nn, "ku")) {
      pa = &KuDsc;
      tt = "diffusion coefficient in u-direction (m2/s)"; }
    else if (!strcmp(nn, "kv")) {
      pa = &KvDsc;
      tt = "diffusion coefficient in v-direction (m2/s)"; }
    else if (!strcmp(nn, "kw")) {
      pa = &KwDsc;
      tt = "diffusion coefficient in w-direction (m2/s)"; }
    else if (!strcmp(nn, "vd")) {
      pa = &VdDsc;
      tt = "deposition velocity (m/s)"; }
    else if (!strcmp(nn, "pd")) {
      pa = &PdDsc;
      tt = "deposition probability"; }
    else if (!strcmp(nn, "li")) {
      pa = &LiDsc;
      tt = "interception length (m)"; }
    else                                                            eX(1);

    if (strstr("bp+bm", nn) && (pa->start))                         eX(10);
    AryFree(pa);
    if (!strchr(FN, PATHSEP))  strcpy(fname, WorkPath);
    else  *fname = 0;
    strcpy(s, FN);
    if (!*s) {
      if (RC > 0)  sprintf(s, "%s%04d", nn, RC);
      else         strcpy(s, nn);
      }
    strcat(fname, s);
    vMsg("%s: reading file %s\n", nn, fname);
    TxtClr(&uhdr);
    TxtClr(&shdr);
    DmnRead(fname, &uhdr, &shdr, pa);                               eG(2);
    if      (strstr("bp", nn)) {
      if (AryAssert(pa, sizeof(float), 2, 0, NX, 0, NY))            eX(3);
      }
    else if (strstr("bm+vd+pd", nn)) {
      if (AryAssert(pa, sizeof(float), 2, 1, NX, 1, NY))            eX(4);
      }
    else if (strstr("vx", nn)) {
      if (AryAssert(pa, sizeof(float), 3, 0, NX, 1, NY, 1, NZ))     eX(5);
      }
    else if (strstr("vy", nn)) {
      if (AryAssert(pa, sizeof(float), 3, 1, NX, 0, NY, 1, NZ))     eX(6);
      }
    else if (strstr("vs+vv", nn)) {
      if (AryAssert(pa, sizeof(float), 3, 1, NX, 1, NY, 0, NZ))     eX(7);
      }
    else if (strstr("li", nn)) {
      if (AryAssert(pa, sizeof(float), 3, 1, NX, 1, NY, 1, NZ))     eX(8);
      }
    else if (strstr("su+sv+sw+sc+ku+kv+kw+dt", nn)) {
      if (AryAssert(pa, sizeof(float), 3, 0, NX, 0, NY, 0, NZ))     eX(9);
      /* if (!strcmp(nn, "sv"))  ShowArray("sv", pa); */
      }
    pc = MsgTok(pn, &pn, " +\t\r\n", &l);
    }
  return 0;
eX_1:
  eMSG("%s is not a known data type! (%s)\n", nn, ePGMs);
eX_2:
  eMSG("can't read from file %s! (%s)\n", fname, ePGMs);
eX_3: eX_4: eX_5: eX_6: eX_7: eX_8: eX_9:
  eMSG("%s has invalid structure! (%s)\n", nn, ePGMs);
eX_10:
  eMSG("duplicate definition of surface! (%s)\n", ePGMs);
  }

/*================================================================== ClearData
*/
int ClearData(          /* Daten lschen        */
char *arg )
  {
  ARYDSC *pa;
  static char *names[] = { "con", "dry", "wet", "icp", "" };
  int n;
  for (n=0; (*names[n]); n++) {
    if      (!strcmp(names[n], "con"))  pa = &ConDsc;
    else if (!strcmp(names[n], "icp"))  pa = &IcpDsc;
    else if (!strcmp(names[n], "dry"))  pa = &DryDsc;
    else if (!strcmp(names[n], "wet"))  pa = &WetDsc;
    else continue;
    if (!pa->start)  continue;
    memset(pa->start, 0, pa->ttlsz);
    }
  if (!strcmp(arg, "all")) {
      n = ClearAllPtl();
      if (MsgVerbose & 0x02)  vMsg("  all particles (%d) removed\n", n);
  }
  T1 = T2;
  return 0;
  }


/*======================================================================= main
*/
void Help( void )
  {
  MsgVerbose = 1;
  vMsg("\nIBJparticle: Particle Dispersion Model, version %s\n", VERSION);
  vMsg("Copyright (C) Dr. Lutz Janicke, Dunum, Germany, 1999-2000\n");
  vMsg("Compiled at %s  %s\n", __DATE__, __TIME__);
  vMsg("IBJparticle comes with ABSOLUTELY NO WARRANTY\n");
  vMsg("IBJparticle is free software under the GNU PUBLIC LICENCE\n\n");
  vMsg("Aufruf: IBJparticle <Pfad> [<Option> ...]\n\n");
  vMsg("<Pfad>  : Name des Arbeitsverzeichnisses\n");
  vMsg(
  "<Option>: -e        -  Echo der Eingabe im Log-File\n"
  "          -h        -  Ausgabe einer Auflistung aller Parameter\n"
  "          -i<Name>  -  Lesen der Eingabe aus File <Name>\n"
  "          -l<Name>  -  Ausgabe der Log-Information in den File <Name>\n"
  "          -q        -  Keine Bildschirmausgabe (quiet)\n"
  "          -r<Zahl>  -  Die Zahl <Zahl> zum Wert von Sd addieren\n"
  "          -T<Zahl>  -  Alle <Zahl> Sekunden Teilchenzhler anzeigen\n"
  "          -v<Zahl>  -  Setzen des verbose levels\n"
  "          -w<Zahl>  -  Die Zahl <Zahl> zum Wert von Wc addieren\n");
  exit(0);
  }

void PrtAbort( int sig_number )
  {
  time_t ti;
  time(&ti);
  MsgVerbose = 1;
  vMsg("%s", ctime(&ti));
  vMsg("IBJparticle aborted (%d)\n", sig_number);
  if (MsgFile) {
    fclose(MsgFile);
    MsgFile = NULL;
    }
  exit(0);
  }


int main(           /* Hauptprogramm, Steuerung der Verfahrensschritte */
int argc,
char *argv[] )      /* Name des Eingabefiles, Optionen                 */
  {
  dP(PrtMain);
  int i, m;
  char *pk, *s, fname[256], cmd[80], arg[200], buf[80];
  char alles[128];
  time_t ti;
  signal(SIGINT, PrtAbort);
  sprintf(alles, "%s", "*c*d*g*p*q*r*w*z*t*l*n*e");
  if (argc < 2)  Help();
  strcpy(CmdName, "prtcmd.txt");
  strcpy(LogName, "prtlog.txt");
  strcpy(ChkName, "prtchk.txt");
  MsgVerbose = 0;
  strcpy(WorkPath, ".");
  for (i=1; i<argc; i++) {
    if (strchr(SYMOPT, *argv[i])) {
      s = argv[i] + 1;
      switch (s[0]) {
        case 'C':  PrtCheck = 1;
                   sscanf(s+1, "%x", &PrtCheck);
                   break;
        case 'e':  PrmEcho = 1;
                   sscanf(s+1, "%x", &PrmEcho);
                   break;
        case 'h':  MsgVerbose = 1;
                   PrmPrint(alles);
                   exit(0);
        case 'i':  strcpy(CmdName, s+1);
                   break;
        case 'l':  strcpy(LogName, s+1);
                   break;
        case 'q':  MsgQuiet = 1;
                   sscanf(s+1, "%x", &MsgQuiet);
                   break;
        case 'r':  sscanf(s+1, "%d", &AddRandom);
                   break;
        case 'T':  Ticks = 1;
                   sscanf(s+1, "%ld", &Ticks);
                   break;
        case 'v':  MsgVerbose = 1;
                   sscanf(s+1, "%x", &MsgVerbose);
                   break;
        case 'w':  sscanf(s+1, "%d", &AddWrite);
                   break;
        default:   printf("unknown option %s\n", argv[i]);
        }
      }
    else strcpy(WorkPath, argv[i]);
    }
  for (pk=WorkPath; (*pk); pk++)  if (*pk == ALTPSEP)  *pk = PATHSEP;
  if ((pk[-1]!=':') && (pk[-1]!=PATHSEP)) {
    pk[0] = PATHSEP;
    pk[+1] = 0;
    }
  if (*LogName != '-') {
    strcpy(fname, MakeFileName(LogName, ".txt"));             eG(26);
    MsgFile = fopen(fname, "wb");  if (!MsgFile)              eX(27);
    }
  m = MsgVerbose;  MsgVerbose = 1;
  time(&ti);
  vMsg("\nIBJparticle: Particle Dispersion Model, version %s\n", VERSION);
  vMsg("Copyright (C) Dr. Lutz Janicke, Dunum, Germany, 1999-2000\n");
  vMsg("IBJparticle comes with ABSOLUTELY NO WARRANTY\n");
  vMsg("IBJparticle is free software under the GNU PUBLIC LICENCE\n\n");
  fprintf(MsgFile, "Compiled at %s  %s\n", __DATE__, __TIME__);
  vMsg("STARTING at %s", ctime(&ti));
  if (MsgFile)
    for (i=0; i<argc; i++)  fprintf(MsgFile, ">%s\n", argv[i]);
  MsgVerbose = m;
  if (PrtCheck) {
      strcpy(fname, MakeFileName(ChkName, ".txt"));
      ChkFile = fopen(fname, "wb");
  }
  GrdInit();
  time(&ti);
  *cmd = 0;
  *arg = 0;

    strcpy(fname, MakeFileName(CmdName, ".txt"));           eG(1);
    while (PrmRead(fname, cmd, arg)) {                      eG(2);
        *fname = 0;
        switch (tolower(cmd[1])) {
            case '*':   fprintf(MsgFile, "%s\n", arg);
                        break;
            case 'c':   ClearData(arg);             eG(25);
                        break;
            case 'd':   ;
                        break;
            case 'e':   ;
                        break;
            case 'g':   DefineGrid();               eG(30);
                        break;
            case 'l':
            case 'n':   ;
                        break;
            case 'p':   ClcModField();              eG(5);
                        break;
            case 'q':   ;
                        break;
            case 'r':   *FN = 0;
                        if (RC >= 0)  RC++;
                        ReadData(arg);              eG(29);
                        break;
            case 't':  if (NULL == strcmp(arg, "*"))
                         PrmPrint(alles);
                       else
                         for (i=0; i<strlen(arg); i++) {
                           strcpy(buf, "*");
                           if (arg[i] != '*')  strncat(buf, arg+i, 1);
                           PrmPrint(buf);
                          }
                        break;
            case 'w':   if (WC >= 0)  WC++;
                        WC += AW;
                        WriteData(arg);             eG(23);
                        AW = 0;
                        break;
            case 'z':   ClcCon(arg);                eG(9);
                        break;
            case '?':   printf("%s, press [RETURN]\n", arg);
                        fgets(buf, 80, stdin);
                        break;
            default:                                eX(11);
        }
    if (tolower(cmd[1] == 'e'))  break;
    }
    PrmRead(NULL, cmd, arg);

  time(&ti);
  MsgVerbose = 1;
  vMsg("FINISHED at %s\n", ctime(&ti));
  return 0;
eX_1: eX_26:
  eMSG("invalid file name! (%s)\n", ePGMs);
eX_2:
  eMSG("can't read data for command %s (%s)\n", cmd, ePGMs);
eX_5:
  eMSG("can't calculate met-fields! (%s)\n", ePGMs);
eX_9:
  eMSG("can't calculate concentration! (%s)\n", ePGMs);
eX_11:
  eMSG("unknown command  >>>%s<<<! (%s)\n", cmd, ePGMs);
eX_23:
  eMSG("can't write %s! (%s)\n", arg, ePGMs);
eX_25:
  eMSG("can't clear data %s! (%s)\n", arg, ePGMs);
eX_27:
  eMSG("can't open log file %s! (%s)\n", fname, ePGMs);
eX_29:
  eMSG("can't read %s! (%s)\n", arg, ePGMs);
eX_30:
  eMSG("can't define grid! (%s)\n", ePGMs);
  }

/*==========================================================================*/
