//======================================================== IBJprt.h
//
// Lagrangian Particle Dispersion Model
// ====================================
//
// Copyright (C) Dr. Lutz Janicke, Dunum, Germany, 1999
// email: lj@janicke.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// history:
//
// last change: 98-07-31  lj
//
//==================================================================

#ifndef  IBJPRT_INCLUDE
#define  IBJPRT_INCLUDE

#ifndef     IBJARY_INCLUDE
  #include "IBJary.h"
#endif

#ifndef     IBJMTH_INCLUDE
  #include "IBJmth.h"
#endif

#ifndef  GRDLOCREC_TYPE
#define  GRDLOCREC_TYPE
typedef struct {
  float x, y, z, s;
  float a, b, c;
  int   i, j, k;
  int flag;
  }  GRDLOCREC;
#endif

#define  PRT_BOT  0x01
#define  PRT_TOP  0x02
#define  PRT_INB  0x04
#define  PRT_OUT  0x10

#define  PRT_PEX        0x0001
#define  PRT_PEY        0x0002
#define  PRT_TRY        0x0004
#define  PRT_3DM        0x0010
#define  PRT_VELCOR     0x0020
#define  PRT_USE_VV     0x0040
#define  PRT_NODRIFT    0x0100
#define  PRT_DEFGRD     0x1000

#define  PRT_CHKPTL  0x0001
#define  PRT_CHKINI  0x0002
#define  PRT_CHKLOC  0x0004
#define  PRT_CHKLST  0x0008
#define  PRT_CHKMOD  0x0010
#define  PRT_CHKTRC  0x0020
#define  PRT_CHKOUT  0x0040
#define  PRT_CHKTRL  0x0080
#define  PRT_CHKREG  0x0100

#define  PRTPERX    (PrtMode & PRT_PEX)
#define  PRTPERY    (PrtMode & PRT_PEY)
#define  PRTRETRY   (PrtMode & PRT_TRY)
#define  PRTMET3    (PrtMode & PRT_3DM)
#define  PRTVELCOR  (PrtMode & PRT_VELCOR)
#define  PRTFLIP  	(PrtMode & PRT_FLIP)
#define  PRTUSE_VV  (PrtMode & PRT_USE_VV)

/*-------------------------------------------------------------------------*/

typedef struct ptlrec *PTLRECPTR;
typedef struct ptlrec { /* Struktur mit den Parametern eines Teilchens  */
  PTLRECPTR next;       /* Pointer auf nchstes Teilchen                */
  PTLRECPTR prev;       /* Pointer auf vorgehendes Teilchen             */
  VECTOR r;             /* Koordinaten des Teilchens [m]                */
  VECTOR v;             /* Geschwindigkeit des Teilchens [m/s]          */
  VECTOR u;             /* Aufgeprgte Geschwindigkeit [m/s]            */
  double t0, t;         /* Laufende Zeit [s]                            */
  int f, n;             /* Flags und Identifizierungsnummer             */
  float g0, g;          /* Masse des Teilchens                          */
  long rnd;
  } PTLREC;

typedef struct {        /* Struktur zur Lokalisierung eines Teilchens   */
  GRDLOCREC m;          /* Position im Rechengitter                     */
  struct {              /* Auszhl-Gitter:                              */
    int i, j, k;        /* Netz-Indizes                                 */
    } s;
  } LOCREC;

typedef struct {    /* Struktur fr Werte des Modell-Feldes     */
  float tau;        /* Zeitschritt [s]                          */
  VECTOR wind;      /* Windgeschwindigkeit                      */
  VECTOR drift;     /* Korrektur-Term                           */
  MATRIX psi;       /* Gedchtnis-Faktor                        */
  MATRIX lam;       /* Stochastischer Anteil                    */
  } MODREC;

extern int PrtMode, PrtCheck;
extern ARYDSC WwDsc, SsDsc, LlDsc, KkDsc, PpDsc, EeDsc;
extern FILE *ChkFile;

#endif  /*##################################################################*/
